(* gamutViewControllerD.ml     View controller delegate for Gamut
 *
 * Copyright (c) 2015 Psellos   http://psellos.com
 *
 * Licensed under the MIT license:
 *     http://www.opensource.org/licenses/mit-license.php
 *)

let frnd f =
    floor (f +. 0.5)

let hsva_of_hsla (h, s, l, a) =
    (* Convert HSLA to HSVA color representation.
     *)
    let os =
        if l <= 0.5 then
            (2.0 *. s *. l) /. (l +. s *. l)
        else
            (2.0 *. s *. (1.0 -. l)) /. (l +. s *. (1.0 -. l))
    in
    let ov =
        if l <= 0.5 then
            l +. s *. l
        else
            l +. s *. (1.0 -. l)
    in
        (h, os, ov, a)


(* What kind of touch right now, based on where the touch *started*.
 *)
type touchtype =
    | TouchNone    (* No touch at the moment *)
    | TouchCircle  (* Touching the circle in the middle *)
    | TouchBack    (* Touching the background *)

class t =
object (self)
    inherit Wrappee.t

    (* Main (only) view.
     *)
    val mutable theView: UiView.t option = None

    (* The circle in the middle of the view.
     *)
    val mutable theCircle : UiBezierPath.t option = None

    (* The text in the circle.
     *)
    val label = "\xce\xa8" (* Greek capital psi *)
    val mutable labelsize = (0.0, 0.0)
    val mutable theLabelfont : UiFont.t option = None

    (* Times control lightness of colors. fgtime tracks seconds since
     * startup (incremented at each timer tick). bgtime is a record of
     * the most recent synchronization of the background and foreground
     * times, but is otherwise static.
     *)
    val mutable fgtime: float = 0.0
    val mutable bgtime: float = 0.0

    (* State of touches.
     *)
    val mutable touchty: touchtype = TouchNone (* What kind of touch now *)
    val mutable touchpoint: Cocoa.point =      (* Most recent touched spot *)
        (0.0, 0.0)

    (* Initial HSL.
     *)
    val ini_hue = 0.5333 (* Initial touch y -> hue *)
    val ini_sat = 0.96   (* Initial touch x -> sat *)
    val ini_lig = 9.85   (* Initial bgtime -> lightness *)

    (* View controller events.
     *)
    method viewController'viewDidAppear'
            (vc: UiViewController.t) (animated: bool) =
        theView <- Some vc#view;
        theCircle <- Some (UiBezierPath.bezierPath ());
        let (_, _, w, h) = vc#view#frame in
        touchpoint <- (ini_sat *. w, ini_hue *. h);
        bgtime <- ini_lig

    method viewController'viewWillDisappear'
            (vc: UiViewController.t) (animated: bool) =
        ()

    method viewControllerShouldAutorotate' (vc: UiViewController.t) =
        false

    (* Timer events.
     *)
    method timerTick' (timer: NsTimer.t) =
        fgtime <- fgtime +. Gamut.interval;
        self#display

    (* Touch events.
     *)
    method view'touchesBegan' (view: UiView.t) (touch: Cocoa.point) =
        (match theCircle with
        | None -> touchty <- TouchBack
        | Some c ->
            touchty <-
                if c#containsPoint' touch then TouchCircle else TouchBack
        );
        if touchty <> TouchCircle then
            touchpoint <- touch

    method view'touchesMoved' (view: UiView.t) (touch: Cocoa.point) =
        if touchty <> TouchCircle then
            touchpoint <- touch

    method view'touchesEnded' (view: UiView.t) (touch: Cocoa.point) =
        if touchty = TouchCircle then
            bgtime <- fgtime;
        touchty <- TouchNone

    method view'touchesCancelled' (view: UiView.t) (touch: Cocoa.point) =
        if touchty = TouchCircle then
            bgtime <- fgtime;
        touchty <- TouchNone

    (* Draw method.
     *)
    method view'drawRect' (v: UiView.t) (r: Cocoa.rect) =
        let white = (0.0, 0.0, 1.0, 1.0) in
        let kernfact = -0.1 in (* Move to the left a little *)
        let basefact = 0.08 in (* Move down a little *)
        let (_, _, vw, vh) = v#frame in
        let (fgcolor, _, resting) = self#colors_of_time fgtime in
        let bgtime' = if touchty = TouchCircle then fgtime else bgtime in
        let (bgcolor, bordcolor, _) = self#colors_of_time bgtime' in
        let labelfont = self#labelfont vw vh in
        let (lw, lh) = self#labelsize labelfont in
        UiKit.set bgcolor;
        UiKit.rectFill (0.0, 0.0, vw, vh);
        self#set_circle vw vh;
        (match theCircle with
        | None -> ()
        | Some c ->
            UiKit.set fgcolor;
            c#fill;
            UiKit.set (if resting then white else bordcolor);
            c#stroke;
        );
        UiKit.set bgcolor;
        let pt =
            (frnd ((vw -. lw) /. 2.0) +. kernfact *. lw,
             frnd ((vh -. lh) /. 2.0) +. basefact *. lw)
        in
            UiKit.string'drawAtPoint'withFont' label pt labelfont

    (* Private methods.
     *)

    method private display =
        match theView with
        | None -> ()
        | Some v -> v#setNeedsDisplay

    method private set_circle w h =
        match theCircle with
        | None -> ()
        | Some bzp ->
            let pi = 4.0 *. atan 1.0 in
            let centerx = frnd (w /. 2.0) in
            let centery = frnd (h /. 2.0) in
            let radius = frnd (w /. 6.0) in
            bzp#removeAllPoints;
            bzp#addArcWithCenter'radius'startAngle'endAngle'clockwise'
                (centerx, centery) radius 0.0 (2.0 *. pi) false;
            bzp#setLineWidth' (frnd (w /. 100.0))

    method private colors_of_time time =
        (* Translate the current touch position and the given time into
         * an interesting color and a contrasting color and an indicator
         * of the resting period (when the color rests at base lightness
         * of 0.5). We use HSLA internally, but return a color in HSVA
         * format.
         *)
        let (lightness, resting) = self#lightness_of_time time in
        let (ih, is, il, ia) as interior =
            match theView with
            | None -> (0.0, 0.0, 0.0, 1.0)   (* Not really possible *)
            | Some view ->
                let (_, _, vw, vh) = view#frame in
                let (x, y) = touchpoint in
                let h = y /. vh in
                let s = x /. vw in
                let l = lightness in
                (h, s, l, 1.0)
        in
        let border =
            if il > 0.5 then (ih, is, il -. 0.2, ia)
            else (ih, is, il +. 0.2, ia)
        in
        (hsva_of_hsla interior, hsva_of_hsla border, resting)

    method private lightness_of_time time =
        (* Return a lightness value (between 0.0 and 1.0) corresponding
         * to the given time, and a boolean saying whether we're in a
         * "rest" phase at the base lightness level (0.5).
         *
         * We want to alternate between light and dark with some fairly
         * longish period, with most of the time spent in the
         * interesting lightnesses between (say) 0.2 and 0.8. There's
         * also a rest phase at the base level (0.5).
         *)
        let phases = [
            (2.0, 0.8); (* Rest phase is first *)
            (0.5, 0.9); (0.5, 1.0); (0.5, 0.9); (0.5, 0.8);
            (2.5, 0.2); (0.5, 0.0); (0.5, 0.2); (2.5, 0.5);
        ]
        in
        let total = List.fold_left (fun n (a, _) -> n +. a) 0.0 phases in
        let btime = mod_float time total in
        let rec go (prev, first) time phases =
            match phases with
            | [] -> (prev, first)
            | (seconds, level) :: _ when time <= seconds ->
                (prev +. (time /. seconds) *. (level -. prev), first)
            | (seconds, level) :: rest ->
                go (level, false) (time -. seconds) rest
        in
        go (snd (List.hd phases), true) btime phases

    method private labelfont vw vh =
        match theLabelfont with
        | Some f -> f
        | None ->
            let size = frnd (vw /. 4.0) in
            let f = UiFont.fontWithName'size' "Helvetica-Oblique" size in
            theLabelfont <- Some f;
            f

    method private labelsize labelfont =
        if labelsize = (0.0, 0.0) then
            labelsize <- UiKit.string'sizeWithFont' label labelfont;
        labelsize

    method private view'drawAppIconRect' (v: UiView.t) (r: Cocoa.rect) =
        (* Draw an app icon at size 60 x 60. Called during development,
         * not during normal use.
         *)
        let side = 60.0 in
        let pside = side *. 1.5 in (* Proportions of icon are different *)
        let white = (0.0, 0.0, 1.0, 1.0) in
        let kernfact = -0.1 in (* Move to the left a little *)
        let basefact = 0.08 in (* Move down a little *)
        let (_, _, vw, vh) = v#frame in
        let (fgcolor, _, resting) = self#colors_of_time 0.0 in
        let (bgcolor, bordcolor, _) = self#colors_of_time ini_lig in
        let labelfont = self#labelfont pside pside in
        let (lw, lh) = self#labelsize labelfont in
        UiKit.set bgcolor;
        UiKit.rectFill (0.0, 0.0, vw, vh);
        self#set_circle pside (pside *. 2.0);
        (match theCircle with
        | None -> ()
        | Some c ->
            UiKit.set fgcolor;
            c#fill;
            UiKit.set (if resting then white else bordcolor);
            c#stroke;
        );
        UiKit.set bgcolor;
        let pt =
            (frnd ((pside -. lw) /. 2.0) +. kernfact *. lw,
             frnd ((pside *. 2.0 -. lh) /. 2.0) +. basefact *. lw)
        in
            UiKit.string'drawAtPoint'withFont' label pt labelfont

end


let () =
    let wrapped robjcv =
        let t = new t in let () = t#setContainer robjcv in t
    in
    Callback.register "GamutViewControllerD.wrapped" wrapped
