/* wrap.m     Simple custom wrappers for Gamut
 *
 * Copyright (c) 2015 Psellos   http://psellos.com
 * Licensed under the MIT license:
 *     http://www.opensource.org/licenses/mit-license.php
 */
# import <stdlib.h>
# import <objc/runtime.h>
# import <UIKit/UIKit.h>
# import "caml/memory.h"
# import "caml/alloc.h"
# import "caml/callback.h"
# import "wrap.h"

/* Represent raw ObjC object pointers in OCaml as nativeint values.
 */
static value wrapRawObjC(id obj)
{
    CAMLparam0();
    CAMLreturn(caml_copy_nativeint((intnat) obj));
}

static id unwrapRawObjC(value objcval)
{
    return (__bridge id) (void *) Nativeint_val(objcval);
}


/* Some key ObjC objects are wrapped up for access from OCaml. To keep
 * things simple, we enter wrapped ObjC objects into an NSMapTable and
 * never remove them. The table causes the objects to be persistent, and
 * it also lets us map from ObjC objects (pointers) to the OCaml objects
 * that wrap them (values).
 *
 * To stay on good terms with the OCaml GC, the OCaml values must be
 * reachable from a global root. For simplicity, we just make them all
 * global roots individually.
 * 
 * In this simple example these few ObjC objects are never destroyed.
 * In a more dynamic setting, we would need to gain control when the
 * OCaml object is GCed. Then we would remove the entry from the table,
 * which would cause (or allow) it to be deallocated on the ObjC side.
 */
static NSMapTable *g_objc_to_ocaml;

static void init_objc_to_ocaml()
{
    if (g_objc_to_ocaml)
        return;
    NSMapTableOptions kopt =
        NSMapTableStrongMemory | NSMapTableObjectPointerPersonality;
    NSMapTableOptions vopt =
        NSMapTableStrongMemory;
    g_objc_to_ocaml =
        [NSMapTable mapTableWithKeyOptions: kopt valueOptions: vopt];
}


static value find_ocaml_wrapper(id obj) {
    init_objc_to_ocaml();
    NSValue *v = [g_objc_to_ocaml objectForKey: obj];
    if (v == nil)
        return 0;
    return * (value *) [v pointerValue];
}


static value wrapObjC(char *wrapcs, id obj)
{
    /* Create a new OCaml object that wraps the given ObjC object.
     * wrapcs gives the name of the closure for creating an OCaml
     * wrapper of the correct type. A closure is registered by each
     * OCaml class that wraps an ObjC class. Each of these OCaml classes
     * inherits from Wrapper.t.
     */
    CAMLparam0();
    CAMLlocal1(val);
    value *valp;
    value *closure;

    closure = (value *) caml_named_value(wrapcs);
    valp = (value *) malloc(sizeof *valp);
    if(closure == NULL || valp == NULL) {
        /* If this happens, things will go very bad from here.
         */
        fprintf(stderr, "wrapObjC failure for %s\n", wrapcs);
        fflush(stderr);
        CAMLreturn(Val_unit); /* Not type safe, should never happen */
    }
    val = caml_callback(*closure, wrapRawObjC(obj));
    *valp = val;
    NSValue *v = [NSValue valueWithPointer: valp];

    /* Enter the wrapping into the table.
     */
    init_objc_to_ocaml();
    [g_objc_to_ocaml setObject: v forKey: obj];
    caml_register_global_root(valp);

    CAMLreturn(val);
}


static value wrapObjC_dyn(id obj)
{
    /* Wrap as wrapObjC(), but obtain the wrapping function name by
     * querying the class name of the object. Be willing to use the
     * wrapping function of a superclass if there is one.
     *
     * Note: this function isn't used right now, but it seems
     * potentially cool.
     */
    Class cl;
    const char *classname;
    char closurename[128];
#   define WRAP ".wrap"

    for (cl = object_getClass(obj); cl != nil; cl = class_getSuperclass(cl)) {
        classname = class_getName(cl);
        if(strlen(classname) + sizeof WRAP > sizeof closurename) {
            fprintf(stderr, "wrapObjC_dyn: class name too long: %s\n",
                    classname);
            fflush(stderr);
            return Val_unit;
        }
        strcpy(closurename, classname);
        strcat(closurename, WRAP);
        if (caml_named_value(closurename) != NULL)
            return wrapObjC(closurename, obj);
    }
    fprintf(stderr, "wrapObjC_dyn: lookup failure for %s\n", closurename);
    fflush(stderr);
    return Val_unit;
}


static id unwrapObjC(value obj) {
    /* Return the ObjC object wrapped by the given OCaml object.
     */
    CAMLparam1(obj);
    CAMLlocal1(objcval);

    objcval =
        caml_callback(
            caml_get_public_method(obj, caml_hash_variant("contents")),
            obj);
    CAMLreturnT(id, unwrapRawObjC(objcval));
}


/* Subclasses of WrapOCaml wrap OCaml objects as ObjC objects.
 */
@implementation WrapOCaml {
    value _contents;
}

- (value) contents
{
    return _contents;
}

- (void) setContents: (value) aValue
{
    _contents = aValue;
}

- (WrapOCaml *) init
{
    /* Create a new OCaml object that is wrapped by self. We match the
     * OCaml class to our own class by name--simple but effective. We
     * then call the closure for creating a wrapped object, which is
     * registered by each OCaml class that can be wrapped in an ObjC
     * class. Such OCaml classes inherit from Wrappee.t.
     */
    const char *classname;
    char closurename[128];
    value *closure;
#   define WRAPPED ".wrapped"

    if((self = [super init]) != nil) {
        classname = object_getClassName(self);
        if(strlen(classname) + sizeof WRAPPED > sizeof closurename) {
            fprintf(stderr, "[WrapOCaml init]: class name too long: %s\n",
                    classname);
            fflush(stderr);
            return nil;
        }
        strcpy(closurename, classname);
        strcat(closurename, WRAPPED);
        closure = (value *) caml_named_value(closurename);
        if(closure == NULL) {
            fprintf(stderr, "[WrapOCaml init]: lookup failure for %s\n",
                    closurename);
            fflush(stderr);
            return nil;
        }
        _contents = caml_callback(*closure, wrapRawObjC(self));
        caml_register_global_root(&_contents);
    }
    return self;
}

- (void) dealloc
{
    if(_contents != 0) {
        caml_remove_global_root(&_contents);
        _contents = 0;
    }
}

@end

/* Conversion utilities.
 */

static void Point_val(CGPoint *ptp, value pointval)
{
    CAMLparam1(pointval);

    ptp->x = Double_val(Field(pointval, 0));
    ptp->y = Double_val(Field(pointval, 1));

    CAMLreturn0;
}

static value Val_rect(CGRect *rectp)
{
    CAMLparam0();
    CAMLlocal1(rectval);

    rectval = caml_alloc(4, 0);
    Store_field(rectval, 0, caml_copy_double(rectp->origin.x));
    Store_field(rectval, 1, caml_copy_double(rectp->origin.y));
    Store_field(rectval, 2, caml_copy_double(rectp->size.width));
    Store_field(rectval, 3, caml_copy_double(rectp->size.height));

    CAMLreturn(rectval);
}

static void Rect_val(CGRect *rectp, value rectval)
{
    CAMLparam1(rectval);

    rectp->origin.x = Double_val(Field(rectval, 0));
    rectp->origin.y = Double_val(Field(rectval, 1));
    rectp->size.width = Double_val(Field(rectval, 2));
    rectp->size.height = Double_val(Field(rectval, 3));

    CAMLreturn0;
}


/* OCaml objects accessed from ObjC.
 *
 * In OCaml these inherit from Wrappee.t. In ObjC they are subclasses of
 * WrapOCaml.
 */

/* An ObjC method name often has ':' in it, which makes it an invalid
 * OCaml method name. Translate ':' in ObjC to '\'' in OCaml. Also, of
 * course, an OCaml method name is a single token.
 */

@implementation GamutAppDelegate

@dynamic window;

- (UIWindow *) window
{
    CAMLparam0();
    CAMLlocal2(selfval, windowval);

    selfval = [self contents];
    windowval =
        caml_callback(
            caml_get_public_method(selfval, caml_hash_variant("window")),
            selfval);
    CAMLreturnT(UIWindow *, unwrapObjC(windowval));
}


- (void) setWindow: (UIWindow *) window
{
    CAMLparam0();
    CAMLlocal2(selfval, windowval);

    selfval = [self contents];
    if((windowval = find_ocaml_wrapper(window)) == 0)
        windowval = wrapObjC("UIWindow.wrap", (id) window);
    (void)
        caml_callback2(
            caml_get_public_method(selfval, caml_hash_variant("setWindow'")),
            selfval, windowval);
    CAMLreturn0;
}


- (void) applicationDidFinishLaunching: (UIApplication *) application
{
    CAMLparam0();
    CAMLlocal2(selfval, applval);

    selfval = [self contents];
    if((applval = find_ocaml_wrapper(application)) == 0)
        applval = wrapObjC("UIApplication.wrap", (id) application);
    (void)
        caml_callback2(
            caml_get_public_method(
                selfval,
                caml_hash_variant("applicationDidFinishLaunching'")),
            selfval, applval);
    CAMLreturn0;
}


- (void) applicationDidReceiveMemoryWarning: (UIApplication *) application
{
    CAMLparam0();
    CAMLlocal2(selfval, applval);

    selfval = [self contents];
    if((applval = find_ocaml_wrapper(application)) == 0)
        applval = wrapObjC("UIApplication.wrap", (id) application);
    (void)
        caml_callback2(
            caml_get_public_method(
                selfval,
                caml_hash_variant("applicationDidReceiveMemoryWarning'")),
            selfval, applval);
    CAMLreturn0;
}

- (void) applicationWillResignActive: (UIApplication *) application
{
    CAMLparam0();
    CAMLlocal2(selfval, applval);

    selfval = [self contents];
    if((applval = find_ocaml_wrapper(application)) == 0)
        applval = wrapObjC("UIApplication.wrap", (id) application);
    (void)
        caml_callback2(
            caml_get_public_method(
                selfval,
                caml_hash_variant("applicationWillResignActive'")),
            selfval, applval);
    CAMLreturn0;
}

- (void) applicationDidBecomeActive: (UIApplication *) application
{
    CAMLparam0();
    CAMLlocal2(selfval, applval);

    selfval = [self contents];
    if((applval = find_ocaml_wrapper(application)) == 0)
        applval = wrapObjC("UIApplication.wrap", (id) application);
    (void)
        caml_callback2(
            caml_get_public_method(
                selfval,
                caml_hash_variant("applicationDidBecomeActive'")),
            selfval, applval);
    CAMLreturn0;
}

@end

@implementation GamutViewControllerD

- (void) viewController: (UIViewController *) vc
         viewDidAppear: (BOOL) animated
{
    CAMLparam0();
    CAMLlocal2(selfval, vcval);

    selfval = [self contents];
    if ((vcval = find_ocaml_wrapper(vc)) == 0)
        vcval = wrapObjC("UIViewController.wrap", (id) vc);
    (void)
        caml_callback3(
            caml_get_public_method(
                selfval,
                caml_hash_variant("viewController'viewDidAppear'")),
            selfval, vcval, Val_bool(!!animated));
    CAMLreturn0;
}


- (void) viewController: (UIViewController *) vc
         viewWillDisappear: (BOOL) animated
{
    CAMLparam0();
    CAMLlocal2(selfval, vcval);

    selfval = [self contents];
    if ((vcval = find_ocaml_wrapper(vc)) == 0)
        vcval = wrapObjC("UIViewController.wrap", (id) vc);
    (void)
        caml_callback3(
            caml_get_public_method(
                selfval,
                caml_hash_variant("viewController'viewWillDisappear'")),
            selfval, vcval, Val_bool(!!animated));
    CAMLreturn0;
}


- (BOOL) viewControllerShouldAutorotate: (UIViewController *) vc
{
    CAMLparam0();
    CAMLlocal3(selfval, vcval, resval);

    selfval = [self contents];
    if ((vcval = find_ocaml_wrapper(vc)) == 0)
        vcval = wrapObjC("UIViewController.wrap", (id) vc);
    resval =
        caml_callback2(
            caml_get_public_method(
                selfval,
                caml_hash_variant("viewControllerShouldAutorotate'")),
            selfval, vcval);
    CAMLreturnT(BOOL, Bool_val(resval));
}


static void touch_event(value selfval, UIView *view, NSSet *touches,
                    UIEvent *event, char *ocmeth)
{
    /* For this simple test, we assume there's just one touch, and we
     * represent it by a point.
     */
    CAMLparam1(selfval);
    CAMLlocal2(viewval, pointval);
    UITouch *touch;
    CGPoint location;

    if((viewval = find_ocaml_wrapper(view)) == 0)
        viewval = wrapObjC("UIView.wrap", (id) view);

    if((touch = (UITouch *) [touches anyObject]) == nil)
        CAMLreturn0; /* Not really possible */

    location = [touch locationInView: view];

    pointval = caml_alloc(2, 0);
    Store_field(pointval, 0, caml_copy_double(location.x));
    Store_field(pointval, 1, caml_copy_double(location.y));

    (void)
        caml_callback3(
            caml_get_public_method(
                selfval,
                caml_hash_variant(ocmeth)),
            selfval,
            viewval,
            pointval);
    CAMLreturn0;
}


- (void) view: (UIView *) view
         touchesBegan: (NSSet *) touches
         withEvent: (UIEvent *) event
{
    touch_event([self contents], view, touches, event, "view'touchesBegan'");
}


- (void) view: (UIView *) view
         touchesMoved: (NSSet *) touches
         withEvent: (UIEvent *) event
{
    touch_event([self contents], view, touches, event, "view'touchesMoved'");
}


- (void) view: (UIView *) view
         touchesEnded: (NSSet *) touches
         withEvent: (UIEvent *) event
{
    touch_event([self contents], view, touches, event, "view'touchesEnded'");
}


- (void) view: (UIView *) view
         touchesCancelled: (NSSet *) touches
         withEvent: (UIEvent *) event
{
    touch_event([self contents], view, touches, event,
                "view'touchesCancelled'");
}


- (void) view: (UIView *) view drawRect: (CGRect) rect
{
    CAMLparam0();
    CAMLlocal3(selfval, viewval, rectval);

    selfval = [self contents];

    if((viewval = find_ocaml_wrapper(view)) == 0)
        viewval = wrapObjC("UIView.wrap", (id) view);

    rectval = Val_rect(&rect);

    (void)
        caml_callback3(
            caml_get_public_method(selfval,
                                   caml_hash_variant("view'drawRect'")),
            selfval, viewval, rectval);
    CAMLreturn0;
}


- (void) timerTick: (NSTimer *) aTimer
{
    CAMLparam0();
    CAMLlocal2(selfval, timerval);


    selfval = [self contents];
    if((timerval = find_ocaml_wrapper(aTimer)) == 0)
        /* Can't report a timer tick if the timer isn't known.
         */
        CAMLreturn0;
    (void)
        caml_callback2(
            caml_get_public_method(selfval, caml_hash_variant("timerTick'")),
            selfval, timerval);
    CAMLreturn0;
}

@end


/* ObjC objects accessed from OCaml.
 *
 * In OCaml these inherit from Wrapper.t. In ObjC they are not limited
 * to any particular class.
 */

/* PSDelegator, any ObjC object that delegates to an OCaml object.
 *
 * (Only used in special cases, to keep things simple.)
 */
@protocol PSDelegatorDelegate
- (WrapOCaml *) delegate;
- (void) setDelegate: (WrapOCaml *) delegate;
@end

value PSDelegator_delegate(value objcval)
{
    /* nativeint -> Wrappee.t
     */
    CAMLparam1(objcval);
    id<PSDelegatorDelegate> delegator = unwrapRawObjC(objcval);
    WrapOCaml *delegate = delegator.delegate;
    CAMLreturn(delegate.contents);
}


value PSDelegator_setDelegate_(value objcval, value delegval)
{
    /* nativeint -> Wrappee.t -> unit
     */
    CAMLparam2(objcval, delegval);
    CAMLlocal1(delegwrapper);
    id<PSDelegatorDelegate> delegator = unwrapRawObjC(objcval);
    delegwrapper = 
        caml_callback(
            caml_get_public_method(delegval, caml_hash_variant("container")),
            delegval);
    delegator.delegate = (WrapOCaml *) unwrapRawObjC(delegwrapper);
    CAMLreturn(Val_unit);
}


/* NSTimer
 */

value NSTimer_scheduledTimer5(value timeInterval, value target, value selector,
                    value userInfo, value repeats)
    /* float -> #Wrappee.t -> string -> #Wrappee.t -> bool -> NsTimer.t
     *
     * For simplicity, target must be wrapped by an ObjC object already.
     * It's OK if userInfo isn't wrapped; we treat this as nil.
     */
{
    CAMLparam5(timeInterval, target, selector, userInfo, repeats);
    CAMLlocal2(targro, uiro);

    double ti = Double_val(timeInterval);
    targro = 
        caml_callback(
            caml_get_public_method(target, caml_hash_variant("container")),
            target);
    id targ = unwrapRawObjC(targro);
    if(targ == nil) {
        /* If the target wasn't already wrapped, things will go very bad
         * from here.
         */
        fprintf(stderr, "NSTimer_scheduledTimer5: target not wrapped\n");
        fflush(stderr);
        CAMLreturn(Val_unit); /* Not type safe, should never happen */
    }

    // Caller supplies selector in OCaml form (with '\''). Translate to
    // ObjC form, then translate to selector.
    //
    char *sels = strdup(String_val(selector));
    char *quote = sels;
    while ((quote = strchr(quote, '\'')) != NULL)
        *quote = ':';
    SEL sel = NSSelectorFromString([NSString stringWithUTF8String: sels]);
    free(sels);

    uiro =
        caml_callback(
            caml_get_public_method(userInfo, caml_hash_variant("container")),
            userInfo);
    id ui = unwrapRawObjC(uiro);
    int rep = Bool_val(repeats);

    NSTimer *timer =
        [NSTimer scheduledTimerWithTimeInterval: ti
                 target: targ
                 selector: sel
                 userInfo: ui
                 repeats: rep ];
    CAMLreturn(wrapObjC("NSTimer.wrap", (id) timer));
}


/* UIKit
 */

value UIKit_RectFill(value rectval)
/* Cocoa.rect -> unit */
{
    CAMLparam1(rectval);
    CGRect rect;

    Rect_val(&rect, rectval);
    UIRectFill(rect);

    CAMLreturn(Val_unit);
}


value UIKit_set(value colorval)
/* UiKit.color -> unit */
{
    CAMLparam1(colorval);

    UIColor *color =
        [UIColor colorWithHue: Double_val(Field(colorval, 0))
                 saturation: Double_val(Field(colorval, 1))
                 brightness: Double_val(Field(colorval, 2))
                 alpha: Double_val(Field(colorval, 3))];
    [color set];

    CAMLreturn(Val_unit);
}


value UIKit_setFill(value colorval)
/* UiKit.color -> unit */
{
    CAMLparam1(colorval);

    UIColor *color =
        [UIColor colorWithHue: Double_val(Field(colorval, 0))
                 saturation: Double_val(Field(colorval, 1))
                 brightness: Double_val(Field(colorval, 2))
                 alpha: Double_val(Field(colorval, 3))];
    [color setFill];

    CAMLreturn(Val_unit);
}


value UIKit_setStroke(value colorval)
/* UiKit.color -> unit */
{
    CAMLparam1(colorval);

    UIColor *color =
        [UIColor colorWithHue: Double_val(Field(colorval, 0))
                 saturation: Double_val(Field(colorval, 1))
                 brightness: Double_val(Field(colorval, 2))
                 alpha: Double_val(Field(colorval, 3))];
    [color setStroke];

    CAMLreturn(Val_unit);
}


value UIKit_string_sizeWithFont_(value stringval, value objcval)
/* string -> nativeint -> Cocoa.size */
{
    CAMLparam2(stringval, objcval);
    CAMLlocal1(sizeval);
    NSString *string;
    UIFont *font;
    CGSize size;

    string = [NSString stringWithUTF8String: String_val(stringval)];
    font = unwrapRawObjC(objcval);

    size = [string sizeWithFont: font];

    sizeval = caml_alloc(2, 0);
    Store_field(sizeval, 0, caml_copy_double(size.width));
    Store_field(sizeval, 1, caml_copy_double(size.height));

    CAMLreturn(sizeval);
}

value UIKit_string_drawAtPoint_withFont_(value stringval, value pointval,
                    value objcval)
/* string -> Cocoa.point -> nativeint -> unit */
{
    CAMLparam3(stringval, pointval, objcval);
    NSString *string;
    CGPoint pt;
    UIFont *font;

    string = [NSString stringWithUTF8String: String_val(stringval)];
    Point_val(&pt, pointval);
    font = unwrapRawObjC(objcval);

    [string drawAtPoint: pt withFont: font];

    CAMLreturn(Val_unit);
}


/* UIFont
 */

value UIFont_fontWithName_size_(value nameval, value sizeval)
/* string -> float -> UiFont.t */
{
    CAMLparam2(nameval, sizeval);
    NSString *name;
    CGFloat size;
    UIFont *font;

    name = [NSString stringWithUTF8String: String_val(nameval)];
    size = Double_val(sizeval);
    font = [UIFont fontWithName: name size: size];

    CAMLreturn(wrapObjC("UIFont.wrap", (id) font));
}


/* UIBezierPath
 */

value UIBezierPath_bezierPath(value unitval)
/* unit -> UiBezierPath.t */
{
    CAMLparam1(unitval);
    UIBezierPath *bzp;

    bzp = [UIBezierPath bezierPath];

    CAMLreturn(wrapObjC("UIBezierPath.wrap", (id) bzp));
}

value UIBezierPath_bezierPathWithOvalInRect_(value rectval)
/* Cocoa.rect -> UiBezierPath.t */
{
    CAMLparam1(rectval);
    CGRect rect;
    UIBezierPath *bzp;

    Rect_val(&rect, rectval);
    bzp = [UIBezierPath bezierPathWithOvalInRect: rect];

    CAMLreturn(wrapObjC("UIBezierPath.wrap", (id) bzp));
}

value UIBezierPath_moveToPoint_(value objcval, value pointval)
/* nativeint -> Cocoa.point -> unit */
{
    CAMLparam2(objcval, pointval);
    UIBezierPath *bzp = unwrapRawObjC(objcval);
    CGPoint pt;
    Point_val(&pt, pointval);
    [bzp moveToPoint: pt];
    CAMLreturn(Val_unit);
}

value UIBezierPath_addArcWithCenter_radius_startAngle_endAngle_clockwise_(
    value objcval, value centval, value radval, value sangval, value eangval,
    value clockwval)
/* nativeint -> Cocoa.point -> float -> float -> float -> bool -> unit */
{
    CAMLparam5(objcval, centval, radval, sangval, eangval);
    CAMLxparam1(clockwval);
    UIBezierPath *bzp = unwrapRawObjC(objcval);
    CGPoint center;
    Point_val(&center, centval);
    [bzp addArcWithCenter: center
                   radius: Double_val(radval)
               startAngle: Double_val(sangval)
                 endAngle: Double_val(eangval)
                clockwise: Bool_val(clockwval)];
    CAMLreturn(Val_unit);
}

value UIBezierPath_addArcWithCenter_radius_startAngle_endAngle_clockwise__BY(
    value *argv, int argn)
{
    return UIBezierPath_addArcWithCenter_radius_startAngle_endAngle_clockwise_(
        argv[0], argv[1], argv[2], argv[3], argv[4], argv[5]);
}

value UIBezierPath_closePath(value objcval)
/* nativeint -> unit */
{
    CAMLparam1(objcval);
    UIBezierPath *bzp = unwrapRawObjC(objcval);
    [bzp closePath];
    CAMLreturn(Val_unit);
}

value UIBezierPath_removeAllPoints(value objcval)
/* nativeint -> unit */
{
    CAMLparam1(objcval);
    UIBezierPath *bzp = unwrapRawObjC(objcval);
    [bzp removeAllPoints];
    CAMLreturn(Val_unit);
}

value UIBezierPath_lineWidth(value objcval)
/* nativeint -> float */
{
    CAMLparam1(objcval);

    UIBezierPath *bzp = unwrapRawObjC(objcval);

    CAMLreturn(caml_copy_double([bzp lineWidth]));
}

value UIBezierPath_setLineWidth_(value objcval, value widval)
/* nativeint -> float -> unit */
{
    CAMLparam2(objcval, widval);

    UIBezierPath *bzp = unwrapRawObjC(objcval);
    [bzp setLineWidth: Double_val(widval)];

    CAMLreturn(Val_unit);
}

value UIBezierPath_fill(value objcval)
/* nativeint -> unit */
{
    CAMLparam1(objcval);

    UIBezierPath *bzp = unwrapRawObjC(objcval);
    [bzp fill];
    CAMLreturn(Val_unit);
}

value UIBezierPath_stroke(value objcval)
/* nativeint -> unit */
{
    CAMLparam1(objcval);

    UIBezierPath *bzp = unwrapRawObjC(objcval);
    [bzp stroke];
    CAMLreturn(Val_unit);
}

value UIBezierPath_containsPoint_(value objcval, value pointval)
/* nativeint -> Cocoa.point -> bool */
{
    CAMLparam2(objcval, pointval);
    CGPoint pt;

    UIBezierPath *bzp = unwrapRawObjC(objcval);
    Point_val(&pt, pointval);

    CAMLreturn(Val_bool([bzp containsPoint: pt]));
}


/* UIView
 */

value UIView_frame(value objcval)
/* nativeint -> Cocoa.rect */
{
    CAMLparam1(objcval);
    CAMLlocal1(rectval);
    CGRect rect;

    UIView *view = unwrapRawObjC(objcval);
    rect = [view frame];
    rectval = Val_rect(&rect);
    CAMLreturn(rectval);
}

value UIView_setNeedsDisplay(value objcval)
/* nativeint -> unit */
{
    CAMLparam1(objcval);

    UIView *view = unwrapRawObjC(objcval);
    [view setNeedsDisplay];
    CAMLreturn(Val_unit);
}


/* UIWindow
 */

value UIWindow_rootViewController(value objcval)
{
    /* For simplicity, assume the root view controller is a
     * PSViewConDelegator.
     */
    CAMLparam1(objcval);
    CAMLlocal1(rvcob);

    UIWindow *window = unwrapRawObjC(objcval);
    PSViewConDelegator *rvc =
        (PSViewConDelegator *) [window rootViewController];
    if((rvcob = find_ocaml_wrapper(rvc)) == 0)
        rvcob = wrapObjC("PSViewConDelegator.wrap", rvc);
    CAMLreturn(rvcob);
}


/* UIViewController
 */

value UIViewController_view(value objcval)
/* nativeint -> UiView.t */
{
    CAMLparam1(objcval);
    CAMLlocal1(viewob);

    UIViewController *vc = unwrapRawObjC(objcval);
    UIView *view = (UIView *) [vc view];
    if((viewob = find_ocaml_wrapper(view)) == 0)
        viewob = wrapObjC("UIView.wrap", (id) view);
    CAMLreturn(viewob);
}


/* UIApplication
 */

value UIApplication_delegate(value objcval)
{
    /* (The Gamut application delegate is coded in OCaml, so it appears
     * in ObjC as a subclass of WrapOCaml.)
     */
    CAMLparam1(objcval);
    UIApplication *appl = unwrapRawObjC(objcval);
    WrapOCaml *deleg = (WrapOCaml *) [appl delegate];
    CAMLreturn([deleg contents]);
}


value UIApplication_sharedApplication(value unitval)
{
    CAMLparam1(unitval);
    CAMLlocal1(applval);
    UIApplication *application = [UIApplication sharedApplication];
    if((applval = find_ocaml_wrapper(application)) == 0)
        applval = wrapObjC("UIApplication.wrap", (id) application);
    CAMLreturn(applval);
}
